import configparser
import logging
import sys
import os

from youtube2usb.file_helper import *
from youtube2usb.messagebox import *


# create a module-level logger object
logger = logging.getLogger(__name__)


class Config:
    """Manage the configuration values of the the application"""

    _CONFIG_FILE = os.path.dirname(__file__) + "/" + "static/config.ini"

    _config: configparser.ConfigParser

    def __init__(self):
        """Initialize new Config object"""
        self._config = configparser.ConfigParser()

        print("Config path: " + self._CONFIG_FILE)

        # read the configuration file
        self._config.read(self._CONFIG_FILE)

        # this will be false if the config file could not be read
        if not test_config_file(self._config):
            # display a message box and log the error
            logger.error("Could not read the configuration file!")
            showerror("Error", "Could not read the configuration file!")

            sys.exit(1)

    def _update_config_file(self):
        """write the changes into the config file"""
        with open(self._CONFIG_FILE, 'w') as configfile:
            self._config.write(configfile)

        logger.info("Updated the config file")

    def get_output_dir(self) -> str:
        """Get the directory where the output files should be placed

        Returns:
            The output directory

        """

        output_dir = self._config['SETTINGS']['OutputDir']

        # no output directory is defined
        if len(output_dir) == 0:
            # display a message box and log the error
            logger.error("Output directory not defined!")
            showerror("Error", "Output directory not defined!")
            
            sys.exit(1)

        # output directory does not exist
        if not directory_exist(output_dir):
            # try to create the output directory
            if not create_directory(output_dir):
                # display a message box and log the error
                logger.error("Output directory could not be created!")
                showerror("Error", "Output directory could not be created!")

                sys.exit(1)

        return output_dir

    def set_output_dir(self, output_dir):
        """Set the directory where the output files should be placed

        Args:
            output_dir (str): The output directory

        """

        self._config['SETTINGS']['OutputDir'] = output_dir
        self._update_config_file()

    def get_database_dir(self) -> str:
        """Get the directory of the database

        Returns:
            The database directory

        """

        database_dir = self._config['SETTINGS']['DatabaseDir']

        database_path = get_file_path(database_dir)

        # database directory does not exist
        if not directory_exist(database_path):
            # try to create the output directory
            if not create_directory(database_path):
                # display a message box and log the error
                logger.error("Database directory could not be created!")
                showerror("Error", "Database directory could not be created!")

                sys.exit(1)

        return database_dir

    def set_database_dir(self, database_dir):
        """Set the directory of the database

        Args:
        database_dir (str): The database directory

        """

        self._config['SETTINGS']['DatabaseDir'] = database_dir
        self._update_config_file()

    def get_playlist_url(self) -> str:
        """Get the url of the playlist

        Returns:
            The url of the playlist

        """

        return self._config['PLAYLIST']['PlaylistUrl']

    def set_playlist_url(self, url):
        """Set the url of the playlist

        Args:
            url (str): The url of the playlist

        """

        self._config['PLAYLIST']['PlaylistUrl'] = url
        self._update_config_file()

    def get_playlist_title(self) -> str:
        """Get the title of the playlist

        Returns:
            The title of the playlist

        """

        return self._config['PLAYLIST'].get('PlaylistTitle', "")

    def set_playlist_title(self, title):
        """Set the title of the playlist

        Args:
            title (str): Title of the playlist

        """

        self._config['PLAYLIST']['PlaylistTitle'] = title
        self._update_config_file()

    def get_playlist_number_of_titles(self) -> str:
        """Get the number of titles in the playlist

        Returns:
            The number of titles in the playlist

        """

        return self._config['PLAYLIST'].get('PlaylistNumberOfTitles', "")

    def set_playlist_number_of_titles(self, titles):
        """Set the number of the titles in the playlist

        Args:
            titles (str): Number of titles

        """

        self._config['PLAYLIST']['PlaylistNumberOfTitles'] = titles
        self._update_config_file()