from pytube import YouTube

from PySide2.QtCore import QThread, SIGNAL

from youtube2usb.file_handler import FileHandler
from youtube2usb.database import Database
from youtube2usb.mp3_converter import Mp3Converter


class Downloader(QThread):

    _output_dir: str
    _video_urls: list = []

    _file_handler: FileHandler
    _database: Database
    _converter: Mp3Converter

    video_size: int

    def __init__(self, output_dir, database):
        QThread.__init__(self)

        self._output_dir = output_dir
        self._database = database

        self._file_handler = FileHandler(self._output_dir)
        self._converter = Mp3Converter()

        # register the callbacks for the mp3 conversion
        self._converter.register_callback_chunk_size(self.callback_set_chunk_size)
        self._converter.register_callback_progress(self.callback_update_chunk)

    def __del__(self):
        self.wait()

    def run(self):
        self.download_all()

    def download_all(self):
        print("Starting download...")
        self._database.connect()

        i = 0

        for url in self._video_urls:
            try:
                video = YouTube(url=url)
            except Exception as e:
                print(e)
            else:
                # insert the url into the database
                # if the song is new in the database the function will return True
                is_new = self._database.insert_song_link(url)

                # try to read the title from the database to speed up
                if is_new:
                    title = video.title
                    self._database.insert_song_title(link=url, title=title)
                else:
                    title = self._database.get_song_title(url)

                # send signal that a new song is going to be downloaded
                # noinspection PyTypeChecker
                self.emit(SIGNAL('downloading_title(QString, int)'), title, i)

                i = i + 1

                # register callback to update the progressbar
                video.register_on_progress_callback(self.show_progress_bar)

                # register callback to send download finish signal
                video.register_on_complete_callback(self.callback_finished)

                # if the file is already in the list skip the download
                if self._file_handler.file_exist(title):
                    continue

                # always use the video with the best resolution
                audio_stream = video.streams.\
                    filter(progressive=True, subtype='mp4').order_by('resolution').desc().first()

                self.video_size = audio_stream.filesize

                # send the file size to set the maximum of the progress bar
                # noinspection PyTypeChecker
                self.emit(SIGNAL('title_file_size(int)'), self.video_size)

                # download the audio file
                audio_stream.download(self._output_dir)

                # noinspection PyTypeChecker
                self.emit(SIGNAL('converting_title()'))

                # convert the video file to audio file
                self._converter.convert_to_mp3(self._output_dir, title)

                # noinspection PyTypeChecker
                self.emit(SIGNAL('converting_finished()'))

        print("finish")

        # close the connection to the database
        self._database.disconnect()

    def set_video_urls(self, video_urls: list):
        self._video_urls = video_urls

    def show_progress_bar(self, stream, chunk, file_handle, bytes_remaining):
        # noinspection PyTypeChecker
        self.emit(SIGNAL('remaining_bytes(int)'), self.video_size - bytes_remaining)

    def callback_finished(self, stream, file_handle):
        # noinspection PyTypeChecker
        self.emit(SIGNAL('download_finished()'))

    def callback_set_chunk_size(self, chunk_size: int):
        # noinspection PyTypeChecker
        self.emit(SIGNAL('title_chunk_size(int)'), chunk_size)

    def callback_update_chunk(self, chunk: int):
        # noinspection PyTypeChecker
        self.emit(SIGNAL('update_chunk(int)'), chunk)
