import re
import json
from collections import OrderedDict

from pytube import request


class Playlist:
    """Represent a playlist object containing all songs

    Provide functions to get information's about the playlist and the containing
    videos and video urls

    Args:
        url (str): The url of the playlist. Defaults to an empty string
        populate (bool): Construct the video urls for all videos. Defaults to False

    """

    _playlist_url: str
    _video_urls: list = []

    _video_name_dict: dict = {}

    def __init__(self, url="", populate=False):
        if len(url) > 0:
            self.set_playlist_url(url)

        if populate:
            self.populate_playlist_urls()

    def set_playlist_url(self, url):
        """Set the url of the playlist

        Args:
            url (str): The url of the playlist
        """

        self._playlist_url = url

    def populate_playlist_urls(self):
        """Populate the playlist urls

        Generate the urls for all videos in the playlist
        """
        self.__populate_video_urls()

    def get_playlist_title(self) -> str:
        """Get the title of the playlist

        Returns:
             The title of the playlist

        """

        return self.__title()

    def get_video_urls(self) -> list:
        """Get the video urls

        Returns:
            List of all video urls in the playlist

        """

        return self._video_urls

    def number_of_titles(self) -> int:
        """Get the number of titles

        Returns:
            The number of titles in the playlist

        """

        return len(self._video_urls)

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    #                                                                             #
    #                   Functions copied from the pytube library                  #
    #                    only from the contrib/playlist.py file                   #
    #                                                                             #
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    @staticmethod
    def __load_more_url(req):
        """
        Given an html page or a fragment thereof, looks for
        and returns the "load more" url if found.
        """

        try:
            load_more_url = 'https://www.youtube.com' + re.search(
                r'data-uix-load-more-href=\"(/browse_ajax\?'
                'action_continuation=.*?)\"', req,
            ).group(1)
        except AttributeError:
            load_more_url = ''
        return load_more_url

    def __parse_links(self):
        """
        Parse the video links from the page source, extracts and
        returns the /watch?v= part from video link href
        It's an alternative for BeautifulSoup
        """

        url = self.__construct_playlist_url()
        req = request.get(url)

        # split the page source by line and process each line
        content = [x for x in req.split('\n') if 'pl-video-title-link' in x]
        link_list = [x.split('href="', 1)[1].split('&', 1)[0] for x in content]

        # The above only returns 100 or fewer links
        # Simulating a browser request for the load more link
        load_more_url = self.__load_more_url(req)
        while len(load_more_url):  # there is an url found
            req = request.get(load_more_url)
            load_more = json.loads(req)
            videos = re.findall(
                r'href=\"(/watch\?v=[\w-]*)',
                load_more['content_html'],
            )
            # remove duplicates
            link_list.extend(list(OrderedDict.fromkeys(videos)))
            load_more_url = self.__load_more_url(
                load_more['load_more_widget_html'],
            )

        return link_list

    def __populate_video_urls(self):
        """
        Construct complete links of all the videos in playlist and
        populate video_urls list
        :return: urls -> string
        """

        base_url = 'https://www.youtube.com'
        link_list = self.__parse_links()

        for video_id in link_list:
            complete_url = base_url + video_id
            self._video_urls.append(complete_url)

    def __title(self):
        """ return playlist title (name) """
        try:
            url = self.__construct_playlist_url()
            req = request.get(url)
            open_tag = '<title>'
            end_tag = '</title>'
            matchresult = re.compile(open_tag + '(.+?)' + end_tag)
            matchresult = matchresult.search(req).group()
            matchresult = matchresult.replace(open_tag, '')
            matchresult = matchresult.replace(end_tag, '')
            matchresult = matchresult.replace('- YouTube', '')
            matchresult = matchresult.strip()

            return matchresult
        except Exception as e:
            print(e)
            return None

    def __construct_playlist_url(self):
        """
        There are two kinds of playlist urls in YouTube. One that contains
        watch?v= in URL, another one contains the "playlist?list=" portion. It
        is preferable to work with the later one.

        :return: playlist url
        """

        if 'watch?v=' in self._playlist_url:
            base_url = 'https://www.youtube.com/playlist?list='
            playlist_code = self._playlist_url.split('&list=')[1]
            return base_url + playlist_code

        # url is already in the desired format, so just return it
        return self._playlist_url
