from time import sleep

from PySide2 import QtWidgets
from PySide2.QtCore import Qt, QObject, SIGNAL
from PySide2.QtWidgets import QProgressDialog, QListWidgetItem

from .. import Database
from .. import Config
from .. import Playlist
from .. import Downloader

from .ui_mainwindow import Ui_MainWindow
from .song_list_widget import SongListWidget

from youtube2usb.ui.dialogs import PlaylistUrlDialog
from youtube2usb.ui.dialogs import OutputPathDialog


class MainWindow(QtWidgets.QMainWindow):

    config: Config
    playlist: Playlist
    downloader: Downloader
    database: Database

    progress_dialog: QProgressDialog

    active_step: int
    number_of_steps: int
    
    current_song_widget: SongListWidget

    def __init__(self):
        super(MainWindow, self).__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        # connect the dialogs
        self.ui.pushButton_playlist_url.clicked.connect(lambda: PlaylistUrlDialog(self))
        self.ui.pushButton_output_path.clicked.connect(lambda: OutputPathDialog(self))

        # connect the download button
        self.ui.pushButton.clicked.connect(lambda: self.downloader.start())

        # init some things like config or playlist
        self.init()

        # connect the signals
        self.__connect_signals()

        # show the main window
        self.show()

        # TODO
        # just a simple test how the list widget is working
        # the code should be moved and extended to a working version
        for url in self.playlist.get_video_urls():
            item = QListWidgetItem(self.ui.listWidget)
            self.ui.listWidget.addItem(item)

            row = SongListWidget(url)
            item.setSizeHint(row.minimumSizeHint())

            self.ui.listWidget.setItemWidget(item, row)

    def __connect_signals(self):
        # noinspection PyTypeChecker
        self.connect(self.downloader, SIGNAL('downloading_title(QString, int)'), self.signal_download_title)

        # noinspection PyTypeChecker
        self.connect(self.downloader, SIGNAL('title_file_size(int)'), self.signal_download_set_filesize)

        # noinspection PyTypeChecker
        self.connect(self.downloader, SIGNAL('remaining_bytes(int)'), self.signal_download_status_update)

        # noinspection PyTypeChecker
        self.connect(self.downloader, SIGNAL('download_finished()'), self.signal_download_finished)

        # noinspection PyTypeChecker
        self.connect(self.downloader, SIGNAL('converting_title()'), self.signal_converting_title)

        # noinspection PyTypeChecker
        self.connect(self.downloader, SIGNAL('title_chunk_size(int)'), self.signal_converting_set_chunk_size)

        # noinspection PyTypeChecker
        self.connect(self.downloader, SIGNAL('update_chunk(int)'), self.signal_converting_status_update)

        # noinspection PyTypeChecker
        self.connect(self.downloader, SIGNAL('converting_finished()'), self.signal_converting_finished)

    def init(self):
        self.active_step = 0
        self.number_of_steps = 7

        # initialize the progressbar dialog
        self._setup_progressbar_dialog()

        # step 1 and 2 are really fast so that the progressbar is not showing
        # a short sleep slow down the step speed and the progressbar will show earlier
        # TODO maybe there is a better way to achieve this behavior

        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        #    Step 1                                               #
        #    Initialize the config module                         #
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        self.config = Config()

        sleep(0.4)
        self._update_step()

        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        #    Step 2                                               #
        #    Read config data                                     #
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        playlist_url = self.config.get_playlist_url()
        output_dir = self.config.get_output_dir()
        database_dir = self.config.get_database_dir()

        sleep(0.4)
        self._update_step()

        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        #    Step 3                                               #
        #    Create new database object                           #
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        self.database = Database(database_dir=database_dir)
        self.database.connect()         # TODO handle the return value
        self.database.create_tables()   # TODO handle the return value
        self.database.disconnect()

        self._update_step()

        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        #    Step 4                                               #
        #    Create new playlist object and populate urls         #
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        self.playlist = Playlist(url=playlist_url, populate=True)

        self._update_step()

        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        #    Step 5                                               #
        #    Write the playlist title and url into config         #
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        self.config.set_playlist_title(self.playlist.get_playlist_title())
        self.config.set_playlist_number_of_titles(str(self.playlist.number_of_titles()))

        self._update_step()

        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        #    Step 6                                               #
        #    Create new downloader object                         #
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        self.downloader = Downloader(output_dir=output_dir, database=self.database)
        self.downloader.set_video_urls(self.playlist.get_video_urls())

        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        #    Step 7                                               #
        #    Update the ui                                        #
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        self._set_playlist_info()

        self.progress_dialog.setValue(self.number_of_steps)

    def _set_playlist_info(self):
        # set the playlist title and number of songs
        self.ui.label_playlist_url.setText(self.playlist.get_playlist_title())
        self.ui.label_playlist_songs.setText(str(self.playlist.number_of_titles()))

    def _cancel_init(self):
        self.close()

    def _update_step(self):
        # increment the step counter and update the progress bar
        self.active_step = self.active_step + 1
        self.progress_dialog.setValue(self.active_step)

    def _setup_progressbar_dialog(self):
        # create a progressbar dialog
        self.progress_dialog = QProgressDialog("Initializing...", "Cancel",
                                               self.active_step,
                                               self.number_of_steps,
                                               self)

        # set the title of the progressbar dialog and a fixed width
        # there should be no reason to resize the dialog so a fixed width could be used
        self.progress_dialog.setWindowTitle("Youtube 2 USB Downloader")
        self.progress_dialog.setFixedWidth(400)

        # make the progressbar modal to lock the code until all steps are finished
        # the minimum duration forces the dialog to show even if the init routine is really fast
        self.progress_dialog.setWindowModality(Qt.WindowModal)
        self.progress_dialog.setMinimumDuration(1)

        # connect cancel function to cancel the initialization
        # cancel will close the program
        # noinspection PyTypeChecker
        QObject.connect(self.progress_dialog, SIGNAL("canceled()"), self._cancel_init())

    def signal_download_title(self, title: str, position: int):
        """Signal that a new title will be downloaded

        :param title: Title of the song
        :type title: str
        :param position: Position in the song list
        :type position: int
        """

        item = self.ui.listWidget.item(position)
        self.current_song_widget = self.ui.listWidget.itemWidget(item)

        self.current_song_widget.label.setText(title)
        self.current_song_widget.label_status.setText("Downloading...")

        # show the progress bar and status label
        self.current_song_widget.progress_bar.show()
        self.current_song_widget.label_status.show()

    def signal_download_set_filesize(self, file_size: int):
        """Signal to set the max of the progressbar to song file size

        :param file_size: Size of the song file
        :type file_size: int
        """

        self.current_song_widget.progress_bar.setMaximum(file_size)

    def signal_download_status_update(self, downloaded_bytes: int):
        """Signal to update the progressbar with the current downloaded bytes

        :param downloaded_bytes: Downloaded bytes of the song
        :type downloaded_bytes: int
        """

        self.current_song_widget.progress_bar.setValue(downloaded_bytes)

    def signal_download_finished(self):
        """Signal that the download is finished"""
        self.current_song_widget.label_status.setText("Finished")

        # hide the progress bar
        self.current_song_widget.progress_bar.hide()

    def signal_converting_title(self):
        """Signal that the converting of the current title has stared"""
        self.current_song_widget.label_status.setText("Converting...")

        # show the progress bar and status label
        self.current_song_widget.progress_bar.show()
        self.current_song_widget.label_status.show()

    def signal_converting_set_chunk_size(self, chunk_size: int):
        """Signal to set the chunk size for converting progress

        :param chunk_size: Chunk size
        :type chunk_size: int
        """

        self.current_song_widget.progress_bar.setMaximum(chunk_size)

    def signal_converting_status_update(self, chunk: int):
        """Signal to update the progressbar with the current downloaded chunks

        :param chunk: Downloaded chunk size
        :type chunk: int
        """

        self.current_song_widget.progress_bar.setValue(chunk)

    def signal_converting_finished(self):
        """Signal that the converting is finished"""
        self.current_song_widget.label_status.setText("Finished")

        # hide the progress bar
        self.current_song_widget.progress_bar.hide()
